'use strict';

import { showDiv, hideDiv, ElementById, parseDateTime, formatDuration } from './helpers.js';
import { getCache, addCache, parseName, ConsoleLog, t } from './app_functions.js';
import { dataStorage } from './app.js';
import { setupCanvas } from './canvas.js';
import { get } from './api.js';
import { go } from './router.js';
import { html, render } from './lit-html.js';
import { showAlertModal } from './overlays.js';

let data;

const journeysTemplate = (data) => html`
	<div class="journeys">
		<header id="header">
			<a class="back icon-back" href="#/" title="${t('back')}">${t('back')}</a>
			<h3>${t('from')}: ${parseName(data.params.fromPoint)}</h3>
			<h3>${t('to')}: ${parseName(data.params.toPoint)}</h3>
		</header>

		${dataStorage.settings.enableCanvas ? html`
			<div id="journeysCanvas">
				<canvas id="canvas"></canvas>
			</div>
		` : html`
			<a class="loadMore icon-arrow2 flipped" title="{{LABEL_EARLIER}}" @click=${() => moreJourneys(data.reqId, 'earlier')}"></a>

			<table>
				<thead>
					<tr>
						<th>${t('departure')}</th>
						<th>${t('arrival')}</th>
						<th>${t('duration')}</th>
						<th>${t('changes')}</th>
						<th>${t('products')}</th>
						<th></th>
					</tr>
				</thead>
				<tbody>
					${Object.keys(data.journeys).sort((a, b) => a - b).map(key => journeyOverviewTemplate(data, key))}
				</tbody>
			</table>

			<a class="loadMore icon-arrow2" title="{{LABEL_LATER}}" @click=${() => moreJourneys(data.reqId, "later")}></a>
		`}
	</div>
`;

const journeyOverviewTemplate = (data, key) => {
	let departure       = data.journeys[key].legs[0].departure;
	let arrival         = data.journeys[key].legs[data.journeys[key].legs.length - 1].arrival;
	let changes         = 0;
	let products        = [];
	let changesDuration = 0;
	let cancelled       = false;

	let departureTime = departure.prognosedTime ? departure.prognosedTime : departure.plannedTime;
	let arrivalTime   = arrival.prognosedTime   ? arrival.prognosedTime   : arrival.plannedTime;
	let duration      = arrivalTime - departureTime;

	for (let legKey in data.journeys[key].legs) {
		let leg = data.journeys[key].legs[legKey];
		if (leg.cancelled !== false) cancelled = true;
		if (leg.isWalking || leg.isTransfer) continue;

		changes = changes+1;
		products.push(leg.line.productName);
	}

	products = [...new Set(products)];

	return html`
	<tr @click=${() => go('/'+data.reqId + '/' + key)}">
		<td><span>${timeTemplate(departure, 'departure')}</span></td>
		<td><span>${timeTemplate(arrival, 'arrival')}</span></td>
		<td title="${changesDuration > 0 ? 'Davon '+formatDuration(changesDuration)+' Umstiegsdauer' : ''}"><span>${formatDuration(duration)}</span></td>
		<td><span>${changes-1}</span></td>
		<td><span>${products.join(', ')}</span></td>
		<td><a class="details-button icon-arrow3"></a></td>
	</tr>`;
}

export const journeysView = async (match) => {
	const reqId = match[0];

	data = getCache('journeys', reqId);

	if (!data) {
		data = await get("/journeys", {"reqId": reqId});
		addCache('journeys', data);
	}

	render(journeysTemplate(data), ElementById('content'))

	if (dataStorage.settings.enableCanvas) {
		setupCanvas(data);
	}
};

const timeTemplate = (data, mode) => {
	let delay = 0;

	if (data.prognosedTime !== null) {
		delay = (data.prognosedTime - data.plannedTime)/60;
	}

	return html`
	${delay > 0 ? html`
		${parseDateTime(data.prognosedTime)} <b>(+${delay})</b>
	` : html`
		${parseDateTime(data.plannedTime)}
	`}`;
}

export const moreJourneys  = async (reqId, mode) => {
	let   data    = getCache('journeys', reqId);
	const newData = await get('/moreJourneys', {"reqId": reqId, "mode": mode});

	data.lastUpdated = newData.lastUpdated; 
	data.journeys    = {...data.journeys, ...newData.journeys};

	addCache('journeys', data);
	journeysView([reqId]);
};
